#!/usr/local/bin/perl -w
#
# compoundSval2.pl version 0.3 
#
# This program preprocesses a Senseval-2 lexical sample file and generates
# output that has compounds detected and words lowercased. The following 
# preprocessing tasks are carried out:
# 
# 1. Convert everything to lowercase.
# 2. Combine all satellite words with their corresponding head words, with 
#    the help of <sats>, </sat>, <head sat id...> tags.
# 3. Form all possible compounds within the contexts.
#
# Copyright (c) 2003
#
# Satanjeev Banerjee, University of Minnesota, Duluth
# bane0025@d.umn.edu
#
# Siddharth Patwardhan, University of Minnesota, Duluth
# patw0006@d.umn.edu
#
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@d.umn.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#
#-----------------------------------------------------------------------------

# We have to use commandline options, so use the necessary package!
use Getopt::Long;

# First check if no commandline options have been provided... in which case
# print out the usage notes!
if ( $#ARGV == -1 )
{
    &minimalUsageNotes();
    exit;
}

# Now get the options!
GetOptions("version", "help", "compfile=s");

# If help has been requested, print out help!
if ( defined $opt_help )
{
    $opt_help = $opt_help;
    &showHelp();
    exit;
}

# If version has been requested, show version!
if ( defined $opt_version )
{
    $opt_version = $opt_version;
    &showVersion();
    exit;
}

# Get the wordsfile
if( defined $opt_compfile)
{
    $wordsFile = $opt_compfile;
}
else
{
    print "Compulsory switch --compfile not provided.\n";
    exit;
}

# Open and hashify the words file
open (WORDS, "$wordsFile") || die ("Couldnt open $wordsFile.\n");
while (<WORDS>)
{
    s/[\r\f\n]//;
    $words{$_} = 1;
}
close WORDS;


# Start preprocessing the input.
$insideContext = 0; 
while (<>)
{
    # Make everything lowercase.
    $_ = lc;

    # Remove everything in square brackets.
#    s/\[.*?\]//g;

    # Remove all special characters represented by &stuff;
#    s/&([a-z0-9]*?;)+//g;
    
    # Get all Satellite words...
    s/<head sats=\"([a-z0-9_\-]*)\..*\">.*<\/head>/<head>$1<\/head>/g;
    s/<head>(.*)\-(.*)<\/head>/<head>$1\'$2<\/head>/g;
    s/<sat.*>.*<\/sat>//g;

    # Identifying the compounds...
    if (!$insideContext)
    {
	if (/<context>/) { $insideContext = 1; }
	else { print; }
    }
    # We allow for the <context> tag to have contextual text on the same line!
    if ($insideContext) 
    {
	$insideContext = 0 if (/<\/context>/);
	print(&compoundify($_));
    }
}

# Subroutine that forms all possible compounds in a given context. It takes as
# input a string (preprocessed context from a senseval-2 lexical sample file) and
# returns a string with all compounds identified and connected by "_".
sub compoundify
{
    my $string = shift;
    my $prefix = ""; # the part upto and including <context>
    my $suffix = ""; # the part include </context> and till the end of the string

    if ($string =~ s/(.*<context>)//) { $prefix = $1; }
    if ($string =~ s/(<\/context>.*)//) { $suffix = $1; }

    # now convert whats left in the string
    my @wordsArray = ();
    
    # remove the <head> </head> tags if present, and remember their "position"
    # ASSUMPTION: <head> and </head> will either both occur or neither occur.
    my $headPosition = -1; # -1 for not present
    my $left = "";
    my $head = "";
    my $right = "";
    if ($string =~ /(.*)<head[^>]*>(.*)<\/head>(.*)/)
    {
	$left = $1;
	$head = $2;
	$right = $3;
	$headPosition = 0; # at LEAST 0!
    }
    else { $left = $string; }

    # get all the words into an array
    $left =~ s/<[^>]*>//g;
    $left =~ s/&(\w*;)+//g;
    $left =~ s/\'//g;
    $left =~ s/\.,/ /g;
    while ($left =~ /(\w+)/g)
    {
	push @wordsArray, $1;
    }

    if ($headPosition == 0)
    {
	$headPosition = $#wordsArray + 1;
	
	$head =~ s/<[^>]*>//g;
	$head =~ s/&(\w*;)+//g;
	while ($head =~ /(\S+)/g)
	{
	    push @wordsArray, $1;
	}
	$right =~ s/<[^>]*>//g;
	$right =~ s/&(\w*;)+//g;
	$right =~ s/\'//g;
	$right =~ s/\.,/ /g;
	while ($right =~ /(\w+)/g)
	{
	    push @wordsArray, $1;
	}	
    }
	
    # now compoundify, GREEDILY!!
    my $firstPointer = 0;
    $string = "";

    while ($firstPointer <= $#wordsArray)
    {
	my $secondPointer = $#wordsArray;
	my $done = 0;
	while ($secondPointer > $firstPointer && !$done)
	{
	    my $temp = join ("_", @wordsArray[$firstPointer..$secondPointer]);
	    if (exists $words{$temp})
	    {
		if ($headPosition > -1 && $firstPointer <= $headPosition && $secondPointer >= $headPosition) 
		{ 
		    $string .= "<head>$temp</head> "; 
		}
		else 
		{ 
		    $string .= "$temp "; 
		}
		$done = 1;
	    }
	    else { $secondPointer--; }
	}
	if (!$done) 
	{ 
	    if ($firstPointer == $headPosition) 
	    { 
		$string .= "<head>$wordsArray[$firstPointer]</head> "; 
	    }
	    else 
	    { 
		$string .= "$wordsArray[$firstPointer] "; 
	    }
	}
	$firstPointer = $secondPointer + 1;
    }

    return ($prefix . $string . $suffix . "\n");
}

# function to output a minimal usage note when the user has not provided any
# commandline options
sub minimalUsageNotes
{
    &printUsage;
    print "Type compoundSval2.pl --help for help.\n";
}

# function to output help messages for this program
sub showHelp
{
    &printUsage;
    print "\nThe program preprocesses Senseval-2 lexical sample files (FILE...)\n";
    print "and generates output that contains SENSEVAL-2 data with compounds\n";
    print "identified and some other preprocessing done. Input can be given at the\n";
    print "standard input device (stdin) or by specifying the files (FILE...) at\n";
    print "the command line. Output is obtained at the standard output device\n";
    print "(stdout).The following preprocessing tasks are carried out:\n";
    print "1. Convert everything to lowercase.\n";
    print "2. Combine all satellite words with their corresponding head words,\n";
    print "   with the help of <sats>, </sat>, <head sat id...> tags.\n";
    print "3. Form all possible compounds within the contexts.\n\n";
    print "OPTIONS:\n";
    print "  --compfile         Compusory switch specifying the file COMPFILE\n";
    print "                     containing the list of all the compound words in \n";
    print "                     WordNet.\n";
    print "  --help             Prints this help message.\n\n";
    print "  --version          Prints the version number of this program.\n\n";
}

# function to output the version number
sub showVersion
{
    print "compoundSval2.pl version 0.3\n";
    print "Copyright (c) 2003 Satanjeev Banerjee, Siddharth Patwardhan & Ted Pedersen.\n";
}

# function to output "ask for help" message when the user's goofed up!
sub printUsage
{
    print "Usage: compoundSval2.pl [{--compfile COMPFILE [FILE...] | --help | --version}]\n";
}
